@@ -18,6 +18,10 @@ from errno import EINVAL, ENOENT, EEXIST
 from itertools import chain, count
 from operator import attrgetter
 from stat import S_ISDIR, S_ISLNK, S_ISREG
+try:
+    from urllib import quote as urlquote, quote as urlquote_from_bytes
+except ImportError:
+    from urllib.parse import quote as urlquote, quote_from_bytes as urlquote_from_bytes
 
 
 supports_symlinks = True
@@ -192,6 +196,18 @@ class _NTFlavour(_Flavour):
             return False
         return parts[-1].partition('.')[0].upper() in self.reserved_names
 
+    def make_uri(self, path):
+        # Under Windows, file URIs use the UTF-8 encoding.
+        drive = path.drive
+        if len(drive) == 2 and drive[1] == ':':
+            # It's a path on a local drive => 'file:///c:/a/b'
+            rest = path.as_posix()[2:].lstrip('/')
+            return 'file:///%s/%s' % (
+                drive, urlquote_from_bytes(rest.encode('utf-8')))
+        else:
+            # It's a path on a network drive => 'file://host/share/a/b'
+            return 'file:' + urlquote_from_bytes(path.as_posix().encode('utf-8'))
+
 
 _NO_FD = None
 
@@ -205,7 +221,16 @@ class _PosixFlavour(_Flavour):
 
     def splitroot(self, part, sep=sep):
         if part and part[0] == sep:
-            return '', sep, part.lstrip(sep)
+            stripped_part = part.lstrip(sep)
+            # According to POSIX path resolution:
+            # http://pubs.opengroup.org/onlinepubs/009695399/basedefs/xbd_chap04.html#tag_04_11
+            # "A pathname that begins with two successive slashes may be
+            # interpreted in an implementation-defined manner, although more
+            # than two leading slashes shall be treated as a single slash".
+            if len(part) - len(stripped_part) == 2:
+                return '', sep * 2, stripped_part
+            else:
+                return '', sep, stripped_part
         else:
             return '', '', part
 
@@ -267,6 +292,12 @@ class _PosixFlavour(_Flavour):
     def is_reserved(self, parts):
         return False
 
+    def make_uri(self, path):
+        # We represent the path using the local filesystem encoding,
+        # for portability to other applications.
+        bpath = path.as_bytes()
+        return 'file://' + urlquote_from_bytes(bpath)
+
 
 _nt_flavour = _NTFlavour()
 _posix_flavour = _PosixFlavour()
@@ -361,6 +392,9 @@ if supports_openat:
 
         rename = _wrap_binary_atfunc(os.rename)
 
+        if sys.version_info >= (3, 3):
+            replace = _wrap_binary_atfunc(os.replace)
+
         def symlink(self, target, pathobj, target_is_directory):
             parent_fd, name = _fdnamepair(pathobj)
             os.symlink(str(target), name, dir_fd=parent_fd)
@@ -482,6 +516,9 @@ class _NormalAccessor(_Accessor):
 
     rename = _wrap_binary_strfunc(os.rename)
 
+    if sys.version_info >= (3, 3):
+        replace = _wrap_binary_strfunc(os.replace)
+
     if nt:
         if supports_symlinks:
             symlink = _wrap_binary_strfunc(os.symlink)
@@ -764,6 +801,8 @@ class PurePath(object):
     def as_bytes(self):
         """Return the bytes representation of the path.  This is only
         recommended to use under Unix."""
+        if sys.version_info < (3, 2):
+            raise NotImplementedError("needs Python 3.2 or later")
         return os.fsencode(str(self))
 
     __bytes__ = as_bytes
@@ -771,6 +810,12 @@ class PurePath(object):
     def __repr__(self):
         return "{}({!r})".format(self.__class__.__name__, str(self))
 
+    def as_uri(self):
+        """Return the path as a 'file' URI."""
+        if not self.is_absolute():
+            raise ValueError("relative path can't be expressed as a file URI")
+        return self._flavour.make_uri(self)
+
     @property
     def _cparts(self):
         # Cached casefolded parts, for hashing and comparison
@@ -781,6 +826,8 @@ class PurePath(object):
             return self._cached_cparts
 
     def __eq__(self, other):
+        if not isinstance(other, PurePath):
+            return NotImplemented
         return self._cparts == other._cparts and self._flavour is other._flavour
 
     def __ne__(self, other):
@@ -794,22 +841,22 @@ class PurePath(object):
             return self._hash
 
     def __lt__(self, other):
-        if self._flavour is not other._flavour:
+        if not isinstance(other, PurePath) or self._flavour is not other._flavour:
             return NotImplemented
         return self._cparts < other._cparts
 
     def __le__(self, other):
-        if self._flavour is not other._flavour:
+        if not isinstance(other, PurePath) or self._flavour is not other._flavour:
             return NotImplemented
         return self._cparts <= other._cparts
 
     def __gt__(self, other):
-        if self._flavour is not other._flavour:
+        if not isinstance(other, PurePath) or self._flavour is not other._flavour:
             return NotImplemented
         return self._cparts > other._cparts
 
     def __ge__(self, other):
-        if self._flavour is not other._flavour:
+        if not isinstance(other, PurePath) or self._flavour is not other._flavour:
             return NotImplemented
         return self._cparts >= other._cparts
 
@@ -820,13 +867,24 @@ class PurePath(object):
                     doc="""The root of the path, if any""")
 
     @property
-    def ext(self):
-        """The final component's extension, if any."""
+    def anchor(self):
+        """The concatenation of the drive and root, or ''."""
+        anchor = self._drv + self._root
+        return anchor
+
+    @property
+    def name(self):
+        """The final path component, if any."""
         parts = self._parts
         if len(parts) == (1 if (self._drv or self._root) else 0):
             return ''
-        basename = parts[-1]
-        if basename == '.':
+        return parts[-1]
+
+    @property
+    def ext(self):
+        """The final component's extension, if any."""
+        basename = self.name
+        if basename == '' or basename == '.':
             return ''
         i = basename.find('.')
         if i == -1:
@@ -1105,7 +1163,7 @@ class Path(PurePath):
     def _raise_closed(self):
         raise ValueError("I/O operation on closed path")
 
-    def _opener(self, name, flags, mode=0o777):
+    def _opener(self, name, flags, mode=0o666):
         # A stub for the opener argument to built-in open()
         return self._accessor.open(self, flags, mode)
 
@@ -1135,7 +1193,7 @@ class Path(PurePath):
     def __getattr__(self, name):
         if name.startswith('st_'):
             return getattr(self._stat, name)
-        return super(Path, self).__getattribute__(name)
+        return self.__getattribute__(name)
 
     def glob(self, pattern):
         """Iterate over this subtree and yield all existing files (of any
@@ -1216,6 +1274,22 @@ class Path(PurePath):
             pass
         return self._stat
 
+    @property
+    def owner(self):
+        """
+        Return the login name of the file owner.
+        """
+        import pwd
+        return pwd.getpwuid(self._stat.st_uid).pw_name
+
+    @property
+    def group(self):
+        """
+        Return the group name of the file gid.
+        """
+        import grp
+        return grp.getgrgid(self._stat.st_gid).gr_name
+
     def raw_open(self, flags, mode=0o777):
         """
         Open the file pointed by this path and return a file descriptor,
@@ -1239,7 +1313,7 @@ class Path(PurePath):
         else:
             return io.open(str(self), mode, buffering, encoding, errors, newline)
 
-    def touch(self, mode=0o777, exist_ok=True):
+    def touch(self, mode=0o666, exist_ok=True):
         """
         Create this file with the given access mode, if it doesn't exist.
         """
@@ -1316,6 +1390,18 @@ class Path(PurePath):
             self._raise_closed()
         self._accessor.rename(self, target)
 
+    def replace(self, target):
+        """
+        Rename this path to the given path, clobbering the existing
+        destination if it exists.
+        """
+        if sys.version_info < (3, 3):
+            raise NotImplementedError("replace() is only available "
+                                      "with Python 3.3 and later")
+        if self._closed:
+            self._raise_closed()
+        self._accessor.replace(self, target)
+
     def symlink_to(self, target, target_is_directory=False):
         """
         Make this path a symlink pointing to the given path.
