@@ -3,8 +3,9 @@ import io
 import os
 import errno
 import pathlib
-import sys
 import shutil
+import stat
+import sys
 import tempfile
 import unittest
 from contextlib import contextmanager
@@ -15,6 +16,11 @@ except ImportError:
     from test import test_support as support
 TESTFN = support.TESTFN
 
+try:
+    import grp, pwd
+except ImportError:
+    grp = pwd = None
+
 
 class _BaseFlavourTest(unittest.TestCase):
 
@@ -62,6 +68,11 @@ class PosixFlavourTest(_BaseFlavourTest):
 
     def test_parse_parts(self):
         check = self._check_parse_parts
+        # Collapsing of excess leading slashes, except for the double-slash
+        # special case.
+        check(['//a', 'b'],             ('', '//', ['//', 'a', 'b']))
+        check(['///a', 'b'],            ('', '/', ['/', 'a', 'b']))
+        check(['////a', 'b'],           ('', '/', ['/', 'a', 'b']))
         # Paths which look like NT paths aren't treated specially
         check(['c:a'],                  ('', '', ['c:a']))
         check(['c:\\a'],                ('', '', ['c:\\a']))
@@ -77,7 +88,10 @@ class PosixFlavourTest(_BaseFlavourTest):
         self.assertEqual(f('/a/b'), ('', '/', 'a/b'))
         self.assertEqual(f('/a/b/'), ('', '/', 'a/b/'))
         # The root is collapsed when there are redundant slashes
-        self.assertEqual(f('//a'), ('', '/', 'a'))
+        # except when there are exactly two leading slashes, which
+        # is a special case in POSIX.
+        self.assertEqual(f('//a'), ('', '//', 'a'))
+        self.assertEqual(f('///a'), ('', '/', 'a'))
         self.assertEqual(f('///a/b'), ('', '/', 'a/b'))
         # Paths which look like NT paths aren't treated specially
         self.assertEqual(f('c:/a/b'), ('', '', 'c:/a/b'))
@@ -147,6 +161,9 @@ class NTFlavourTest(_BaseFlavourTest):
 # Tests for the pure classes
 #
 
+with_fsencode = unittest.skipIf(sys.version_info < (3, 2),
+    'os.fsencode has been introduced in version 3.2')
+
 class _BasePurePathTest(unittest.TestCase):
 
     # keys are canonical paths, values are list of tuples of arguments
@@ -159,7 +176,7 @@ class _BasePurePathTest(unittest.TestCase):
             ('', 'a', 'b'), ('a', '', 'b'), ('a', 'b', ''),
             ],
         '/b/c/d': [
-            ('a', '/b/c', 'd'), ('a', '//b//c', 'd/'),
+            ('a', '/b/c', 'd'), ('a', '///b//c', 'd/'),
             ('/a', '/b/c', 'd'),
             # empty components get removed
             ('/', 'b', '', 'c/d'), ('/', '', 'b/c/d'), ('', '/b/c/d'),
@@ -229,14 +246,21 @@ class _BasePurePathTest(unittest.TestCase):
             self.assertEqual(P(pathstr).as_posix(), pathstr)
         # Other tests for as_posix() are in test_equivalences()
 
-    @unittest.skipIf(sys.version_info < (3, 2),
-                     'os.fsencode has been introduced in version 3.2')
+    @with_fsencode
     def test_as_bytes_common(self):
         sep = os.fsencode(self.sep)
         P = self.cls
         self.assertEqual(P('a/b').as_bytes(), b'a' + sep + b'b')
         self.assertEqual(bytes(P('a/b')), b'a' + sep + b'b')
 
+    @with_fsencode
+    def test_as_uri_common(self):
+        P = self.cls
+        with self.assertRaises(ValueError):
+            P('a').as_uri()
+        with self.assertRaises(ValueError):
+            P().as_uri()
+
     def test_repr_common(self):
         for pathstr in ('a', 'a/b', 'a/b/c', '/', '/a/b', '/a/b/c'):
             p = self.cls(pathstr)
@@ -257,6 +281,9 @@ class _BasePurePathTest(unittest.TestCase):
         self.assertNotEqual(P('a/b'), P())
         self.assertNotEqual(P('/a/b'), P('/'))
         self.assertNotEqual(P(), P('/'))
+        self.assertNotEqual(P(), "")
+        self.assertNotEqual(P(), {})
+        self.assertNotEqual(P(), int)
 
     def test_match_common(self):
         P = self.cls
@@ -318,6 +345,11 @@ class _BasePurePathTest(unittest.TestCase):
         assertLess(a, d)
         assertLess(b, c)
         assertLess(c, d)
+        if sys.version_info > (3,):
+            with self.assertRaises(TypeError):
+                P() < {}
+        else:
+            P() < {}
 
     def test_parts_common(self):
         sep = self.sep
@@ -405,6 +437,25 @@ class _BasePurePathTest(unittest.TestCase):
         self.assertEqual(P('/').root, sep)
         self.assertEqual(P('/a/b').root, sep)
 
+    def test_anchor_common(self):
+        P = self.cls
+        sep = self.sep
+        self.assertEqual(P('').anchor, '')
+        self.assertEqual(P('a/b').anchor, '')
+        self.assertEqual(P('/').anchor, sep)
+        self.assertEqual(P('/a/b').anchor, sep)
+
+    def test_name_common(self):
+        P = self.cls
+        self.assertEqual(P('').name, '')
+        self.assertEqual(P('.').name, '')
+        self.assertEqual(P('/').name, '')
+        self.assertEqual(P('a/b').name, 'b')
+        self.assertEqual(P('/a/b').name, 'b')
+        self.assertEqual(P('/a/b/.').name, 'b')
+        self.assertEqual(P('a/b.py').name, 'b.py')
+        self.assertEqual(P('/a/b.py').name, 'b.py')
+
     def test_ext_common(self):
         P = self.cls
         self.assertEqual(P('').ext, '')
@@ -458,12 +509,26 @@ class PurePosixPathTest(_BasePurePathTest):
 
     def test_root(self):
         P = self.cls
-        # This is an UNC path under Windows
-        self.assertEqual(P('//a/b').root, '/')
+        self.assertEqual(P('/a/b').root, '/')
+        self.assertEqual(P('///a/b').root, '/')
+        # POSIX special case for two leading slashes
+        self.assertEqual(P('//a/b').root, '//')
 
     def test_eq(self):
         P = self.cls
         self.assertNotEqual(P('a/b'), P('A/b'))
+        self.assertEqual(P('/a'), P('///a'))
+        self.assertNotEqual(P('/a'), P('//a'))
+
+    @with_fsencode
+    def test_as_uri(self):
+        from urllib.parse import quote_from_bytes
+        P = self.cls
+        self.assertEqual(P('/').as_uri(), 'file:///')
+        self.assertEqual(P('/a/b.c').as_uri(), 'file:///a/b.c')
+        self.assertEqual(P('/a/b%#c').as_uri(), 'file:///a/b%25%23c')
+        self.assertEqual(P('/a/b\xe9').as_uri(),
+                         'file:///a/b' + quote_from_bytes(os.fsencode('\xe9')))
 
     def test_match(self):
         P = self.cls
@@ -477,6 +542,8 @@ class PurePosixPathTest(_BasePurePathTest):
         self.assertTrue(P('/').is_absolute())
         self.assertTrue(P('/a').is_absolute())
         self.assertTrue(P('/a/b/').is_absolute())
+        self.assertTrue(P('//a').is_absolute())
+        self.assertTrue(P('//a/b').is_absolute())
 
     def test_normcase(self):
         P = self.cls
@@ -491,6 +558,27 @@ class PurePosixPathTest(_BasePurePathTest):
         self.assertIs(False, P('/foo/bar').is_reserved())
         self.assertIs(False, P('/dev/con/PRN/NUL').is_reserved())
 
+    def test_join(self):
+        P = self.cls
+        p = P('//a')
+        pp = p.join('b')
+        self.assertEqual(pp, P('//a/b'))
+        pp = P('/a').join('//c')
+        self.assertEqual(pp, P('//c'))
+        pp = P('//a').join('/c')
+        self.assertEqual(pp, P('/c'))
+
+    def test_getitem_common(self):
+        # Basically the same as join()
+        P = self.cls
+        p = P('//a')
+        pp = p['b']
+        self.assertEqual(pp, P('//a/b'))
+        pp = P('/a')['//c']
+        self.assertEqual(pp, P('//c'))
+        pp = P('//a')['/c']
+        self.assertEqual(pp, P('/c'))
+
 
 class PureNTPathTest(_BasePurePathTest):
     cls = pathlib.PureNTPath
@@ -532,6 +620,24 @@ class PureNTPathTest(_BasePurePathTest):
         self.assertEqual(P('C:a/B'), P('c:A/b'))
         self.assertEqual(P('//Some/SHARE/a/B'), P('//somE/share/A/b'))
 
+    @with_fsencode
+    def test_as_uri(self):
+        from urllib.parse import quote_from_bytes
+        P = self.cls
+        with self.assertRaises(ValueError):
+            P('/a/b').as_uri()
+        with self.assertRaises(ValueError):
+            P('c:a/b').as_uri()
+        self.assertEqual(P('c:/').as_uri(), 'file:///c:/')
+        self.assertEqual(P('c:/a/b.c').as_uri(), 'file:///c:/a/b.c')
+        self.assertEqual(P('c:/a/b%#c').as_uri(), 'file:///c:/a/b%25%23c')
+        self.assertEqual(P('c:/a/b\xe9').as_uri(), 'file:///c:/a/b%C3%A9')
+        self.assertEqual(P('//some/share/').as_uri(), 'file://some/share/')
+        self.assertEqual(P('//some/share/a/b.c').as_uri(),
+                         'file://some/share/a/b.c')
+        self.assertEqual(P('//some/share/a/b%#c\xe9').as_uri(),
+                         'file://some/share/a/b%25%23c%C3%A9')
+
     def test_match_common(self):
         P = self.cls
         # Absolute patterns
@@ -642,6 +748,27 @@ class PureNTPathTest(_BasePurePathTest):
         self.assertEqual(P('//a/b/').root, '\\')
         self.assertEqual(P('//a/b/c/d').root, '\\')
 
+    def test_anchor(self):
+        P = self.cls
+        self.assertEqual(P('c:').anchor, 'c:')
+        self.assertEqual(P('c:a/b').anchor, 'c:')
+        self.assertEqual(P('c:/').anchor, 'c:\\')
+        self.assertEqual(P('c:/a/b/').anchor, 'c:\\')
+        self.assertEqual(P('//a/b').anchor, '\\\\a\\b\\')
+        self.assertEqual(P('//a/b/').anchor, '\\\\a\\b\\')
+        self.assertEqual(P('//a/b/c/d').anchor, '\\\\a\\b\\')
+
+    def test_name(self):
+        P = self.cls
+        self.assertEqual(P('c:').name, '')
+        self.assertEqual(P('c:/').name, '')
+        self.assertEqual(P('c:a/b').name, 'b')
+        self.assertEqual(P('c:/a/b').name, 'b')
+        self.assertEqual(P('c:a/b.py').name, 'b.py')
+        self.assertEqual(P('c:/a/b.py').name, 'b.py')
+        self.assertEqual(P('//My.py/Share.php').name, '')
+        self.assertEqual(P('//My.py/Share.php/a/b').name, 'b')
+
     def test_ext(self):
         P = self.cls
         self.assertEqual(P('c:').ext, '')
@@ -907,7 +1034,7 @@ class _BasePathTest(unittest.TestCase):
         self.assertIs(False, p['foo'].exists())
         self.assertIs(False, P('/xyzzy').exists())
 
-    def test_open(self):
+    def test_open_common(self):
         p = self.cls(BASE)
         with p['fileA'].open('r') as f:
             self.assertIsInstance(f, io.TextIOBase)
@@ -1098,6 +1225,20 @@ class _BasePathTest(unittest.TestCase):
         with self.assertRaises(AttributeError):
             p.foo
 
+    @unittest.skipUnless(pwd, "the pwd module is needed for this test")
+    def test_owner(self):
+        p = self.cls(BASE)['fileA']
+        uid = p.stat().st_uid
+        name = pwd.getpwuid(uid).pw_name
+        self.assertEqual(name, p.owner)
+
+    @unittest.skipUnless(grp, "the grp module is needed for this test")
+    def test_group(self):
+        p = self.cls(BASE)['fileA']
+        gid = p.stat().st_gid
+        name = grp.getgrgid(gid).gr_name
+        self.assertEqual(name, p.group)
+
     def test_unlink(self):
         p = self.cls(BASE)['fileA']
         p.unlink()
@@ -1127,7 +1268,25 @@ class _BasePathTest(unittest.TestCase):
         self.assertEqual(os.stat(r).st_size, size)
         self.assertFileNotFound(q.restat)
 
-    def test_touch(self):
+    def test_replace(self):
+        P = self.cls(BASE)
+        p = P['fileA']
+        if sys.version_info < (3, 3):
+            self.assertRaises(NotImplementedError, p.replace, p)
+            return
+        size = p.stat().st_size
+        # Replacing a non-existing path
+        q = P['dirA', 'fileAA']
+        p.replace(q)
+        self.assertEqual(q.stat().st_size, size)
+        self.assertFileNotFound(p.restat)
+        # Replacing another (existing) path
+        r = rel_join('dirB', 'fileB')
+        q.replace(r)
+        self.assertEqual(os.stat(r).st_size, size)
+        self.assertFileNotFound(q.restat)
+
+    def test_touch_common(self):
         P = self.cls(BASE)
         p = P['newfileA']
         self.assertFalse(p.exists())
@@ -1139,7 +1298,6 @@ class _BasePathTest(unittest.TestCase):
         p.touch(mode=0o700, exist_ok=False)
         self.assertTrue(p.exists())
         self.assertRaises(OSError, p.touch, exist_ok=False)
-        # XXX better test `mode` arg
 
     def test_mkdir(self):
         P = self.cls(BASE)
@@ -1241,6 +1399,35 @@ class PosixPathTest(_BasePathTest):
         with self.assertRaises(ValueError):
             print(path.resolve())
 
+    def test_open_mode(self):
+        old_mask = os.umask(0)
+        self.addCleanup(os.umask, old_mask)
+        p = self.cls(BASE)
+        with p['new_file'].open('wb'):
+            pass
+        st = os.stat(join('new_file'))
+        self.assertEqual(stat.S_IMODE(st.st_mode), 0o666)
+        os.umask(0o022)
+        with p['other_new_file'].open('wb'):
+            pass
+        st = os.stat(join('other_new_file'))
+        self.assertEqual(stat.S_IMODE(st.st_mode), 0o644)
+
+    def test_touch_mode(self):
+        old_mask = os.umask(0)
+        self.addCleanup(os.umask, old_mask)
+        p = self.cls(BASE)
+        p['new_file'].touch()
+        st = os.stat(join('new_file'))
+        self.assertEqual(stat.S_IMODE(st.st_mode), 0o666)
+        os.umask(0o022)
+        p['other_new_file'].touch()
+        st = os.stat(join('other_new_file'))
+        self.assertEqual(stat.S_IMODE(st.st_mode), 0o644)
+        p['masked_new_file'].touch(mode=0o750)
+        st = os.stat(join('masked_new_file'))
+        self.assertEqual(stat.S_IMODE(st.st_mode), 0o750)
+
     @with_symlinks
     def test_resolve_loop(self):
         # Loop detection for broken symlinks under POSIX
